;; Generated package description from iterators.el  -*- no-byte-compile: t -*-
(define-package "iterators" "0.1.1.0.20201201.160709" "Functions for working with iterators" '((emacs "25")) :url "https://elpa.gnu.org/packages/iterators.html" :authors '(("Michael Heerdegen" . "michael_heerdegen@web.de")) :maintainer '("Michael Heerdegen" . "michael_heerdegen@web.de") :keywords '("extensions" "elisp"))
